* {
    margin: 0;
    padding: 0;
    box-sizing: border-box;
}

body {
    font-family: Arial, sans-serif;
    background-color: #1C1c1c;
    color: #eae9e9;
    line-height: 1.6;
    display: flex;
    flex-direction: column;
    justify-content: center;
    align-items: center;
    height: 100vh;
}

header {
    color: #fff;
    padding: 36px 0;
    text-align: center;
}

header h1 {
    font-weight: 400;
}

header h3 {
    font-weight: 100;
}

/* Container to hold buttons side by side */
.button-container {
    display: flex;
    gap: 20px;
}

/* Style for the buttons */
.button-container a {
    width: 200px;
    padding: 15px 25px;
    text-align: center;
    background: linear-gradient(0.40turn, #7900C9 25%, #0088FF 300%);
    color: white;
    text-decoration: none;
    border-radius: 5px;
    font-size: 16px;
}

/* Button hover effect */
.button-container a:hover {
    background: #0088FF;
}