#!/usr/bin/env sh
REAL_PATH="$(readlink -f $0)"
GK_PATH="$(dirname "$(dirname "$(dirname "$REAL_PATH")")")"
CLI="$GK_PATH/resources/app.asar/src/main/static/cli.js"
GITKRAKEN="$GK_PATH/gitkraken"
ELECTRON_RUN_AS_NODE=1 "$GITKRAKEN" "$CLI" "$@"
exit $?
